/*
** Hash-A-Lot by Ben Slusky <sluskyb@paranoiacs.org>
** 
** This program will read a passphrase from standard input and print a binary
** (not printable) hash to standard output. The output is suitable for use as
** an encryption key.
**
** USAGE:
**		hashalot [ -s _salt_ ] _hashtype_
**	OR
**		_hashtype_ [ -s _salt_ ]
**
** Most of the code was cribbed from the kerneli.org patch to util-linux,
** by Marc Mutz <Marc@Mutz.com>. Most of what wasn't, was cribbed from GnuPG,
** v.1.0.3 (http://www.gnupg.org/).
*/

#include <libgen.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/mman.h>

#include "rmd160.h"
#include "sha512.h"

#define HASHBUFFLEN 32
#define PASSWDBUFFLEN 130

typedef int (*phash_func_t)(char dest[HASHBUFFLEN], const char src[], size_t src_len);

static int
phash_rmd160(char dest[HASHBUFFLEN], const char src[], size_t src_len)
{
	char tmp[PASSWDBUFFLEN] = { 'A', 0, };
	char key[RMD160_HASH_SIZE * 2] = { 0, };

	strncpy(tmp + 1, src, PASSWDBUFFLEN - 1);
	tmp[PASSWDBUFFLEN - 1] = '\0';
  
	rmd160_hash_buffer(key, src, src_len);
	rmd160_hash_buffer(key + RMD160_HASH_SIZE, tmp, src_len + 1 /* dangerous! */);

	memcpy(dest, key, HASHBUFFLEN);

	memset (tmp, 0, PASSWDBUFFLEN);        /* paranoia */
	memset (key, 0, RMD160_HASH_SIZE * 2); /* paranoia */

	return HASHBUFFLEN;
}

static int
phash_rmd160compat(char dest[HASHBUFFLEN], const char src[], size_t src_len)
{
	phash_rmd160(dest,src,src_len);
	return 16;
}

static int
phash_sha256(char dest[HASHBUFFLEN], const char src[], size_t src_len)
{
	memset(dest, 0, HASHBUFFLEN);
	sha256_hash_buffer((char *) src, src_len, dest, HASHBUFFLEN);
	return HASHBUFFLEN;
}

static int
phash_sha384(char dest[HASHBUFFLEN], const char src[], size_t src_len)
{
	memset(dest, 0, HASHBUFFLEN);
	sha384_hash_buffer((char *) src, src_len, dest, HASHBUFFLEN);
	return HASHBUFFLEN;
}

static int
phash_sha512(char dest[HASHBUFFLEN], const char src[], size_t src_len)
{
	memset(dest, 0, HASHBUFFLEN);
	sha512_hash_buffer((char *) src, src_len, dest, HASHBUFFLEN);
	return HASHBUFFLEN;
}

struct func_table_t {
	const char *name;
	phash_func_t func;
} static func_table[] = {
	{ "ripemd160",	phash_rmd160 },
	{ "rmd160",	phash_rmd160 },
	{ "rmd160compat", phash_rmd160compat },
	{ "sha256",	phash_sha256 },
	{ "sha384",	phash_sha384 },
	{ "sha512",	phash_sha512 },
	{ 0, 0 }
};

static int
show_usage(const char argv0[])
{
	struct func_table_t *p = func_table;

	fprintf (stderr,
		 "%s: missing or unknown hash type requested\n"
		 "\n"
		 "usage:\n"
		 "    hashalot [ -s SALT ] HASHTYPE\n"
		 "  or\n"
		 "    HASHTYPE [ -s SALT ]\n"
		 "\n"
		 "supported values for HASHTYPE: ", argv0);

	for (; p->name; ++p)
		fprintf (stderr, "%s ", p->name);

	fprintf (stderr, "\n");
	
	return 1;
}

static phash_func_t 
phash_lookup(const char phash_name[])
{
	struct func_table_t *p = func_table;

	if (!phash_name)
		return 0;

	for (; p->name && strcmp(phash_name, p->name); ++p);
  
	return p->func;
}

/* A function to read the passphrase either from the terminal or from
 * an open file descriptor */
static char *
xgetpass(const char *prompt)
{
	if (isatty(STDIN_FILENO))	/* terminal */
		return getpass(prompt); /* FIXME getpass(3) obsolete */
	else {				/* file descriptor */
		char *pass = NULL;
		int buflen, i;

		buflen=0;
		for (i=0; ; i++) {
			if (i >= buflen - 1) {
				/* we're running out of space in the buffer. 
				 * Make it bigger: */
				char *tmppass = pass;
				buflen += 128;
				pass = (char *) realloc(tmppass, buflen);
				if (pass == NULL) {
					/* realloc failed. Stop reading _now_. */
					fprintf(stderr, "not enough memory while reading passphrase\n");
					pass = tmppass; /* the old buffer hasn't changed */
					break;
				}
			}
			if (read(STDIN_FILENO, pass+i, 1) != 1 || pass[i] == '\n')
				break;
		}
		if (pass == NULL)
			return "";
		else {
			pass[i] = 0;
			return pass;
		}
	}
}

static void *
xmalloc (size_t size) {
        void *p;

        if (size == 0)
                return NULL;

        p = malloc(size);
        if (p == NULL) {
		perror("malloc");
                exit(1);
        }

        return p;
}

/* function to append a "salt" to the passphrase, to better resist
 * dictionary attacks */
static char *
salt_passphrase(const char *pass, const char *salt) {
	char *buf = xmalloc(strlen(pass) + strlen(salt) + 1);
	sprintf(buf, "%s%s", pass, salt);
	free(pass);
	return buf;
}

int
main(int argc, char *argv[]) 
{
	char *pass, passhash[HASHBUFFLEN], *salt = NULL;
	phash_func_t func;
	int hashlen, c;

	while ((c = getopt(argc, argv, "s:")) != -1) {
		switch (c) {
                case 's':
                        salt = optarg;
                        break;
                default:
                        show_usage(argv[0]);
                }
	}

	if (!(func = phash_lookup(basename(argv[0])))) 
		/* lookup failed, so try next argv */
		if (!(func = phash_lookup(argv[optind]))) {
		/* lookup failed again */
			show_usage (*argv);
			exit(EXIT_FAILURE);
		} 

	assert (func != 0);

	/* try to lock memory so it doesn't get swapped out for sure */
	if (mlockall(MCL_CURRENT | MCL_FUTURE) == -1) {
		perror("mlockall");
		fputs("Warning: couldn't lock memory, are you root?\n", stderr);
	}

	/* here we acquire the precious passphrase... */
	pass = xgetpass("Enter passphrase: ");
	if (salt)
		pass = salt_passphrase(pass, salt);
	hashlen = func(passhash, pass, strlen(pass));
	memset (pass, 0, strlen (pass)); /* paranoia */

	if (write(STDOUT_FILENO, passhash, hashlen) != hashlen)
		perror("write");

	memset (passhash, 0, HASHBUFFLEN); /* paranoia again */

	exit(EXIT_SUCCESS);
}

